<?php

namespace DebugToolbar\Services;

use ExpressionEngine\Service\Addon\Addon;
use DebugToolbar\Services\XmlService;

class ToolbarService
{
    /**
     * @var array
     */
    protected array $template_groups = [];

    /**
     * The theme to use if no other theme is set
     * @var string
     */
    public string $default_theme = "default";

    /**
     * The available positions for the toolbar to live
     * @var array
     */
    public array $toolbar_positions = [
        'bottom-left',
        'top-left',
        'bottom-right',
        'top-right',
    ];

    /**
     * Wrapper to setup and return the toolbar settings
     */
    public function getSettings()
    {
        if (!isset(ee()->session->cache['eedt']['settings'])) {
            if (ee()->extensions->active_hook('eedt_init_settings') === true) {
                $defaults = ee('eedt:SettingsService')->getDefaults();
                $defaults = ee()->extensions->call('eedt_init_settings', $defaults);
                ee('eedt:SettingsService')->setDefaults($defaults);
            }
            ee()->session->cache['eedt']['settings'] = ee('eedt:SettingsService')->getSettings();
        }

        return ee()->session->cache['eedt']['settings'];
    }

    /**
     * Takes the included files and breaks up into mutli arrays for use in the debugger
     * @param array $files
     * @return array
     */
    public function setupFiles(array $files): array
    {
        sort($files);

        $path_third = realpath(ee('eedt:OutputService')->thirdPartyPath());
        $path_ee = realpath(SYSPATH);
        $path_first_modules = realpath(PATH_MOD);
        $bootstrap_file = FCPATH . SELF;
        $return = [
            'third_party_addon' => [],
            'first_party_modules' => [],
            'bootstrap_file' => [],
            'expressionengine_core' => [],
            'composer' => [],
            'other_files' => [],
        ];

        foreach ($files as $file) {
            $output = str_replace($path_ee, '#system', $file);
            if (strpos($file, 'autoload.php') !== false) {
                $return['composer'][] = $output;
            }

            if (strpos($file, $path_third) === 0) {
                $return['third_party_addon'][] = $output;
                continue;
            }

            if (strpos($file, $path_first_modules) === 0) {
                $return['first_party_modules'][] = $output;
                continue;
            }

            if (strpos($file, $bootstrap_file) === 0) {
                $return['bootstrap_file'] = $output;
                continue;
            }

            if (strpos($file, $path_ee) === 0) {
                $return['expressionengine_core'][] = $output;
                continue;
            }

            $return['other_files'][] = $output;
        }

        return $return;
    }

    public function setupCookies(): array
    {
        $providers = ee('App')->getProviders();
        $consents = $cookies = [];

        //echo ee('ee:CookieRegistry')->loadCookiesSettings();
        foreach ($_COOKIE as $key => $value) {
            if (str_starts_with($key, 'exp_')) {
                $old_key = $key;
                $key = substr($key, 4, strlen($key));
                if (ee('ee:CookieRegistry')->isRegistered($key)) {
                    $type = 'Necessary';
                    if (ee('ee:CookieRegistry')->isTargeting($key)) {
                        $type = 'Targeting';
                    } else if (ee('ee:CookieRegistry')->isPerformance($key)) {
                        $type = 'Performance';
                    } else if (ee('ee:CookieRegistry')->isFunctionality($key)) {
                        $type = 'Functionality';
                    }

                    $consents[$key] = array_merge([
                        'key' => $old_key,
                        'value' => $value,
                        'type' => $type,
                    ], ee('ee:CookieRegistry')->getCookieSettings($key));

                } else {
                    $cookies[$old_key] = $value;
                }

            } else {
                $cookies[$key] = $value;
            }
        }

        return [
            'registered' => $consents,
            'unregistered' => $cookies,
        ];
    }

    /**
     * Wrapper to setup the Database panel SQL queries
     * @return array
     */
    public function setupQueries(): array
    {
        $dbs = [];

        // Let's determine which databases are currently connected to
        foreach (get_object_vars(ee()) as $EE_object) {
            if (is_object($EE_object) && is_subclass_of(get_class($EE_object), 'CI_DB')) {
                $dbs[] = $EE_object;
            }
        }

        $output = [];
        if (count($dbs) == 0) {
            return $output;
        }
        // Load the text helper so we can highlight the SQL
        ee()->load->helper('text');

        $count = 0;
        $total_time = 0;
        foreach ($dbs as $db) {
            $count++;

            if (count($db->queries) != 0) {
                foreach ($db->queries as $key => $val) {
                    $total_time = $total_time + $db->query_times[$key];
                    $time = number_format($db->query_times[$key], 4);
                    $output['queries'][] = ['query' => highlight_code($val, ENT_QUOTES), 'time' => $time];
                }
            }

        }

        $output['total_time'] = number_format($total_time, 4);

        return $output;
    }

    /**
     * Wrapper to setup the benchmark data
     * @return array
     */
    public function setupBenchmarks()
    {
        $profile = [];
        foreach (ee()->benchmark->marker as $key => $val) {
            // We match the "end" marker so that the list ends
            // up in the order that it was defined
            if (preg_match("/(.+?)_end/i", $key, $match)) {
                if (isset(ee()->benchmark->marker[$match[1] . '_end']) and isset(ee()->benchmark->marker[$match[1] . '_start'])) {
                    $profile[$match[1]] = ee()->benchmark->elapsed_time($match[1] . '_start', $key);
                }
            }
        }

        return $profile;
    }

    /**
     * Breaks up the template log data to create the chart
     *
     * @param array $log
     * @return array
     */
    public function formatTmplLog(array $log)
    {
        $return = [];
        foreach ($log as $item) {
            $return[] = [
                'time' => $item['time'],
                'memory' => (float)$item['memory'],
                'desc' => mb_convert_encoding($item['message'], 'UTF-8', 'ISO-8859-1'), //a little sanity for UTF-8
            ];
        }

        return $return;
    }

    /**
     * Returns a JSON string that can be used by the Template Chart JS
     *
     * @param $log array
     * @return string
     */
    public function formatTmplChartJson(array $data): string
    {
        foreach ($data as $key => $value) {
            $data[$key]['desc'] = preg_replace("/&#?[a-z0-9]{2,8};/i", '', $data[$key]['desc']);
            $data[$key]['memory_display'] = ee('eedt:ToolbarService')->filesizeFormat($data[$key]['memory']);
            $data[$key]['time'] = number_format($data[$key]['time'], 4);
        }

        return json_encode($data);
    }

    /**
     * Format a number of bytes into a human readable format.
     * Optionally choose the output format and/or force a particular unit
     *
     * @param int $bytes The number of bytes to format. Must be positive
     * @param string $format Optional. The output format for the string
     * @param string $force Optional. Force a certain unit. B|KB|MB|GB|TB
     * @return  string              The formatted file size
     */
    public function filesizeFormat($val, $digits = 3, $mode = "SI", $bB = "B")
    { //$mode == "SI"|"IEC", $bB == "b"|"B"
        $si = ["", "k", "M", "G", "T", "P", "E", "Z", "Y"];
        $iec = ["", "Ki", "Mi", "Gi", "Ti", "Pi", "Ei", "Zi", "Yi"];
        switch (strtoupper($mode)) {
            case "SI" :
                $factor = 1000;
                $symbols = $si;
                break;
            case "IEC" :
                $factor = 1024;
                $symbols = $iec;
                break;
            default :
                $factor = 1000;
                $symbols = $si;
                break;
        }
        switch ($bB) {
            case "b" :
                $val *= 8;
                break;
            default :
                $bB = "B";
                break;
        }
        for ($i = 0; $i < count($symbols) - 1 && $val >= $factor; $i++) {
            $val /= $factor;
        }
        $p = strpos($val, ".");
        if ($p !== false && $p > $digits) {
            $val = round($val);
        } else if ($p !== false) {
            $val = round($val, $digits - $p);
        }

        return round($val, $digits) . " " . $symbols[$i] . $bB;
    }

    /**
     * Checks the system for the available themes and sets up as $key => $value array
     * @return array
     */
    public function getThemes()
    {
        $path = ee('eedt:OutputService')->themePath() . '/eedt/themes/';
        $d = dir($path);
        $themes = [];
        $bad = ['.', '..'];
        while (false !== ($entry = $d->read())) {
            if (is_dir($path . $entry) && !in_array($entry, $bad)) {
                $name = ucwords(str_replace('_', ' ', $entry));
                $themes[$entry] = $name;
            }
        }
        $d->close();
        return $themes;
    }

    /**
     * Create Theme Asset URLs
     *
     * @param string $theme
     * @return string
     */
    public function createThemeUrl($theme, $sub_dir = '')
    {
        $path = ee('eedt:OutputService')->themePath();
        $url = ee('eedt:OutputService')->themeUrl();
        if (is_dir($path . "eedt/themes/" . $theme . "/$sub_dir/")) {
            return $url . "eedt/themes/" . $theme . "/$sub_dir/";
        }

        return $url . "eedt/themes/" . $this->default_theme . "/$sub_dir/";
    }

    /**
     * Returns the ACT for the given params
     * @param string $class
     * @param string $method
     */
    public function fetchActionId($method, $class)
    {
        ee()->load->dbforge();
        ee()->db->select('action_id');
        $query = ee()->db->get_where('actions', ['class' => $class, 'method' => $method]);
        return $query->row('action_id');
    }

    /**
     * Returns the action URL for the given params
     * @param string $method
     * @param string $class
     */
    public function getActionUrl($method, $class = 'Eedt')
    {
        $url = \site_url();
        return $url . '?ACT=' . $this->fetchActionId($method, $class);
    }

    /**
     * Creates the internal ACT URL for use by extensions
     * @param string $act_method
     * @param string $act_class
     * @return string
     */
    public function createActUrl($act_method, $act_class = 'Eedt')
    {
        return $url = $this->getActionUrl('act') . AMP . 'class=' . $act_class . AMP . 'method=' . $act_method;
    }

    /**
     * Takes the panel data and writes them to $path
     * @param \DebugToolbar\Services\XmlService $xml
     * @param array $panels
     * @param string $path
     * @return void
     */
    public function cachePanels(XmlService $xml, array $panels, string $path): void
    {
        $xml->setRootName('EEDT');
        $xml->initiate();
        $xml->startBranch('panels');
        foreach ($panels as $panel) {
            $xml->startBranch($panel->getName() . '_panel');
            $xml->addNode('name', $panel->getName(), [], true);
            $xml->addNode('data_target', $panel->getTarget(), [], true);
            $xml->addNode('button_icon', $panel->getButtonIcon(), [], true);
            $xml->addNode('button_icon_alt_text', $panel->getButtonIconAltText(), [], true);
            $xml->addNode('button_label', $panel->getButtonLabel(), [], true);
            $xml->addNode('output', base64_encode($panel->getPanelContents()), [], true);
            $xml->endBranch();
        }

        $xml->endBranch();
        $xml = $xml->getXml(false);

        $filename = $path . $this->makeCacheFilename();

        $string = mb_convert_encoding($xml, 'UTF-8', 'ISO-8859-1');
        $gz = gzopen($filename . '.gz', 'w9');
        gzwrite($gz, $string);
        gzclose($gz);

        chmod($filename . '.gz', 0777);
    }

    /**
     * Creates the Toolbar panel cache filename
     * @return string
     */
    public function makeCacheFilename(): string
    {
        return '.' . ee()->session->userdata['session_id'] . '.eedt';
    }

    public function getCachePath(): string
    {
        return PATH_CACHE . '/eedt/';
    }

    /**
     * Builds the JS Config to be output as JSON
     * @param array $vars
     * @return array
     */
    public function jsConfig(array $vars = []): array
    {
        $config = [];
        $config['template_debugging_enabled'] = $vars['template_debugging_enabled'];
        $config['panels'] = [];
        $config['cp'] = ee()->input->get('D') == 'cp' ? true : false;
        $config['base_css_url'] = $vars['theme_css_url'];
        $config['base_js_url'] = $vars['theme_js_url'];
        $config['panel_ajax_url'] = str_replace("&amp;", "&", $this->getActionUrl('act') . AMP);

        foreach ($vars['panels'] as $panel) {
            $config['panels'][] = [
                'name' => $panel->getName(),
                'js' => $panel->getJs(),
                'css' => $panel->getCss(),
                'panel_fetch_url' => $panel->getPanelFetchUrl() ? $panel->getPanelFetchUrl() :
                    str_replace("&amp;", "&", $this->createActUrl("GetPanelData")) . "&panel=" . $panel->getName(),
            ];
        }

        return $config;
    }

    /**
     * Checks to verify the MySQL Query Cache is enabled or not
     * @return string
     */
    public function verifyMysqlQueryCache()
    {
        $data = ee()->db->query("SHOW VARIABLES LIKE 'have_query_cache'")->row_array();
        if (!empty($data['Value']) && $data['Value'] == 'YES') {
            return 'y';
        }
    }

    /**
     * @return array
     */
    public function getTemplateGroups(): array
    {
        if (!$this->template_groups) {
            $groups = ee()->db->select()->from('template_groups')->get()->result_array();
            foreach ($groups as $group) {
                $this->template_groups[$group['group_id']] = $group['group_name'];
            }
        }

        return $this->template_groups;
    }

    /**
     * @return bool
     */
    public function canViewToolbar(): bool
    {
        $settings = $this->getSettings();
        $allowed_roles = $settings['allowed_roles'] ?? [1];
        return in_array(ee()->session->userdata('role_id'), $allowed_roles);
    }

    public function shouldCompileToolbar(): bool
    {
        //override to disable the toolbar from even starting
        if (ee()->input->get('disable_toolbar') == 'yes' ||
            ee()->input->get('C') == 'javascript' ||
            (ee()->input->get('ACT') && ee()->input->get('frontedit') == 'on') ||
            (ee()->input->get('ACT') && ee()->input->get('prolet')) ||
            (strpos(ee()->input->server('REQUEST_URI'), 'themes/ee/pro/js') !== false) || //Pro Edit
            ee()->input->get('modal_form') == 'y' ||
            (ee()->input->get('ui') && ee()->input->get('plugin') == 'markitup') ||
            (ee()->input->get('D') == 'cp' && ee()->input->get('C') == 'jumps') || //jump menu
            (ee()->input->get('ACT') && ee()->input->get('addon') && ee()->input->get('file')) //cp add-on page
        ) {
            return false;
        }

        //If its an AJAX request (eg: EE JS Combo loader or jQuery library load) then call it a day...
        $ignore_tmpl_types = ['js', 'css', 'feed'];
        if (AJAX_REQUEST ||
            (property_exists(ee(), "TMPL") && in_array(ee()->TMPL->template_type, $ignore_tmpl_types)) ||
            (isset(ee()->TMPL->template_type) && in_array(ee()->TMPL->template_type, $ignore_tmpl_types))
        ) {
            return false;
        }

        return true;
    }

    /**
     * @param string $add_on
     * @return bool
     */
    public function isAddonInstalled(string $add_on): bool
    {
        $return = false;
        $provider = ee('Addon')->get($add_on);
        if ($provider instanceof Addon) {
            $return = $provider->isInstalled();
        }

        return $return;
    }
}